if SERVER then return end

include("sh_boneanimlib.lua")

usermessage.Hook("resetluaanim", function(um)
	local ent = um:ReadEntity()
	local anim = um:ReadString()
	local tim = um:ReadFloat()
	if tim == 0 then tim = nil end
	if ent:IsValid() then
		ent:ResetLuaAnimation(anim, tim)
	end
end)

usermessage.Hook("setluaanim", function(um)
	local ent = um:ReadEntity()
	local anim = um:ReadString()
	local tim = um:ReadFloat()
	if tim == 0 then tim = nil end
	if ent:IsValid() then
		ent:SetLuaAnimation(anim, tim)
	end
end)

usermessage.Hook("stopluaanim", function(um)
	local ent = um:ReadEntity()
	local anim = um:ReadString()
	local tim = um:ReadFloat()
	if tim == 0 then tim = nil end
	if ent:IsValid() then
		ent:StopLuaAnimation(anim, tim)
	end
end)

usermessage.Hook("stopluaanimgp", function(um)
	local ent = um:ReadEntity()
	local animgroup = um:ReadString()
	local tim = um:ReadFloat()
	if tim == 0 then tim = nil end
	if ent:IsValid() then
		ent:StopLuaAnimationGroup(animgroup, tim)
	end
end)

usermessage.Hook("stopallluaanim", function(um)
	local ent = um:ReadEntity()
	local tim = um:ReadFloat()
	if tim == 0 then tim = nil end
	if ent:IsValid() then
		ent:StopAllLuaAnimations(tim)
	end
end)

local TYPE_GESTURE = TYPE_GESTURE
local TYPE_POSTURE = TYPE_POSTURE
local TYPE_STANCE = TYPE_STANCE
local TYPE_SEQUENCE = TYPE_SEQUENCE

local Animations = GetLuaAnimations()

local function LuaBuildBonePositions(pl, iNumBones, iNumPhysBones)
	local tLuaAnimations = pl.LuaAnimations
	for sGestureName, tGestureTable in pairs(tLuaAnimations) do
		local iCurFrame = tGestureTable.Frame
		local tFrameData = tGestureTable.FrameData[iCurFrame]
		local fFrameDelta = tGestureTable.FrameDelta
		local fDieTime = tGestureTable.DieTime
		local fPower = tGestureTable.Power
		if fDieTime and fDieTime - 0.1 <= CurTime() then
			fPower = fPower * (fDieTime - CurTime()) * 10
		end
		local fAmount = fPower * fFrameDelta

		if fDieTime and fDieTime <= CurTime() or tGestureTable.ShouldPlay and not tGestureTable.ShouldPlay(pl, sGestureName, tGestureTable, iCurFrame, tFrameData, fFrameDelta, fPower) then
			pl:StopLuaAnimation(sGestureName)
		elseif not tGestureTable.PreCallback or not tGestureTable.PreCallback(pl, sGestureName, tGestureTable, iCurFrame, tFrameData, fFrameDelta) then
			if tGestureTable.Type == TYPE_GESTURE then
				local tPrevFrameData = tGestureTable.PrevFrameData[iCurFrame]
				if tPrevFrameData then
					for iBoneID, tBoneInfo in pairs(tPrevFrameData.BoneInfo) do
						if type(iBoneID) ~= "number" then
							iBoneID = pl:LookupBone(iBoneID)
						end

						local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
						if vCurBonePos then
							local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
							if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
								local vUp = aCurBoneAng:Up()
								local vRight = aCurBoneAng:Right()
								local vForward = aCurBoneAng:Forward()
								mBoneMatrix:Translate(tBoneInfo.MU * fPower * vUp + tBoneInfo.MR * fPower * vRight + tBoneInfo.MF * fPower * vForward)
								mBoneMatrix:Rotate(Angle(tBoneInfo.RR * fPower, tBoneInfo.RU * fPower, tBoneInfo.RF * fPower))
								pl:SetBoneMatrix(iBoneID, mBoneMatrix)
							end
						end
					end
				end

				for iBoneID, tBoneInfo in pairs(tFrameData.BoneInfo) do
					if type(iBoneID) ~= "number" then
						iBoneID = pl:LookupBone(iBoneID)
					end

					local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
					if vCurBonePos then
						local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
						if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
							local vUp = aCurBoneAng:Up()
							local vRight = aCurBoneAng:Right()
							local vForward = aCurBoneAng:Forward()
							mBoneMatrix:Translate(fAmount * tBoneInfo.MU * vUp + fAmount * tBoneInfo.MR * vRight + fAmount * tBoneInfo.MF * vForward)
							mBoneMatrix:Rotate(Angle(fAmount * tBoneInfo.RR, fAmount * tBoneInfo.RU, fAmount * tBoneInfo.RF))
							pl:SetBoneMatrix(iBoneID, mBoneMatrix)
						end
					end
				end

				tGestureTable.FrameDelta = tGestureTable.FrameDelta + FrameTime() * tFrameData.FrameRate
				if tGestureTable.FrameDelta > 1 then
					tGestureTable.Frame = iCurFrame + 1
					tGestureTable.FrameDelta = 0
					if tGestureTable.Frame > #tGestureTable.FrameData then
						pl:StopLuaAnimation(sGestureName)
					end
				end
			elseif tGestureTable.Type == TYPE_POSTURE then
				if fFrameDelta < 1 and tGestureTable.TimeToArrive then
					fFrameDelta = math.min(1, fFrameDelta + FrameTime() * (1 / tGestureTable.TimeToArrive))
					tGestureTable.FrameDelta = fFrameDelta
				end

				for iBoneID, tBoneInfo in pairs(tFrameData.BoneInfo) do
					if type(iBoneID) ~= "number" then
						iBoneID = pl:LookupBone(iBoneID)
					end

					local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
					if vCurBonePos then
						local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
						if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
							local vUp = aCurBoneAng:Up()
							local vRight = aCurBoneAng:Right()
							local vForward = aCurBoneAng:Forward()
							mBoneMatrix:Translate(fAmount * tBoneInfo.MU * vUp + fAmount * tBoneInfo.MR * vRight + fAmount * tBoneInfo.MF * vForward)
							mBoneMatrix:Rotate(Angle(fAmount * tBoneInfo.RR, fAmount * tBoneInfo.RU, fAmount * tBoneInfo.RF))
							pl:SetBoneMatrix(iBoneID, mBoneMatrix)
						end
					end
				end
			elseif tGestureTable.Type == TYPE_STANCE then
				local tPrevFrameData = tGestureTable.PrevFrameData[iCurFrame]
				if tPrevFrameData then
					for iBoneID, tBoneInfo in pairs(tPrevFrameData.BoneInfo) do
						if type(iBoneID) ~= "number" then
							iBoneID = pl:LookupBone(iBoneID)
						end

						local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
						if vCurBonePos then
							local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
							if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
								local vUp = aCurBoneAng:Up()
								local vRight = aCurBoneAng:Right()
								local vForward = aCurBoneAng:Forward()
								mBoneMatrix:Translate(fPower * tBoneInfo.MU * vUp + fPower * tBoneInfo.MR * vRight + fPower * tBoneInfo.MF * vForward)
								mBoneMatrix:Rotate(Angle(fPower * tBoneInfo.RR, fPower * tBoneInfo.RU, fPower * tBoneInfo.RF))
								pl:SetBoneMatrix(iBoneID, mBoneMatrix)
							end
						end
					end
				end

				local fFrameDelta = tGestureTable.FrameDelta
				for iBoneID, tBoneInfo in pairs(tFrameData.BoneInfo) do
					if type(iBoneID) ~= "number" then
						iBoneID = pl:LookupBone(iBoneID)
					end

					local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
					if vCurBonePos then
						local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
						if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
							local vUp = aCurBoneAng:Up()
							local vRight = aCurBoneAng:Right()
							local vForward = aCurBoneAng:Forward()
							mBoneMatrix:Translate(fAmount * tBoneInfo.MU * vUp + fAmount * tBoneInfo.MR * vRight + fAmount * tBoneInfo.MF * vForward)
							mBoneMatrix:Rotate(Angle(fAmount * tBoneInfo.RR, fAmount * tBoneInfo.RU, fAmount * tBoneInfo.RF))
							pl:SetBoneMatrix(iBoneID, mBoneMatrix)
						end
					end
				end

				tGestureTable.FrameDelta = tGestureTable.FrameDelta + FrameTime() * tFrameData.FrameRate
				if tGestureTable.FrameDelta > 1 then
					tGestureTable.Frame = iCurFrame + 1
					tGestureTable.FrameDelta = 0
					if tGestureTable.Frame > #tGestureTable.FrameData then
						tGestureTable.Frame = tGestureTable.RestartFrame or 1
					end
				end
			else
				local tPrevFrameData = tGestureTable.PrevFrameData[iCurFrame]
				if tPrevFrameData then
					for iBoneID, tBoneInfo in pairs(tPrevFrameData.BoneInfo) do
						if type(iBoneID) ~= "number" then
							iBoneID = pl:LookupBone(iBoneID)
						end

						local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
						if vCurBonePos then
							local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
							if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
								local vUp = aCurBoneAng:Up()
								local vRight = aCurBoneAng:Right()
								local vForward = aCurBoneAng:Forward()
								mBoneMatrix:Translate(fPower * tBoneInfo.MU * vUp + fPower * tBoneInfo.MR * vRight + fPower * tBoneInfo.MF * vForward)
								mBoneMatrix:Rotate(Angle(fPower * tBoneInfo.RR, fPower * tBoneInfo.RU, fPower * tBoneInfo.RF))
								pl:SetBoneMatrix(iBoneID, mBoneMatrix)
							end
						end
					end
				end

				local fFrameDelta = tGestureTable.FrameDelta
				for iBoneID, tBoneInfo in pairs(tFrameData.BoneInfo) do
					if type(iBoneID) ~= "number" then
						iBoneID = pl:LookupBone(iBoneID)
					end

					local vCurBonePos, aCurBoneAng = pl:GetBonePosition(iBoneID)
					if vCurBonePos then
						local mBoneMatrix = pl:GetBoneMatrix(iBoneID)
						if not tBoneInfo.Callback or not tBoneInfo.Callback(pl, iNumBones, iNumPhysBones, mBoneMatrix, iBoneID, vCurBonePos, aCurBoneAng, fFrameDelta, fPower) then
							local vUp = aCurBoneAng:Up()
							local vRight = aCurBoneAng:Right()
							local vForward = aCurBoneAng:Forward()
							mBoneMatrix:Translate(fAmount * tBoneInfo.MU * vUp + fAmount * tBoneInfo.MR * vRight + fAmount * tBoneInfo.MF * vForward)
							mBoneMatrix:Rotate(Angle(fAmount * tBoneInfo.RR, fAmount * tBoneInfo.RU, fAmount * tBoneInfo.RF))
							pl:SetBoneMatrix(iBoneID, mBoneMatrix)
						end
					end
				end

				tGestureTable.FrameDelta = tGestureTable.FrameDelta + FrameTime() * tFrameData.FrameRate
				if tGestureTable.FrameDelta > 1 then
					tGestureTable.Frame = iCurFrame + 1
					tGestureTable.FrameDelta = 0
					if tGestureTable.Frame > #tGestureTable.FrameData then
						tGestureTable.Frame = tGestureTable.RestartFrame or 1
					end
				end
			end

			if tGestureTable.Callback then
				tGestureTable.Callback(pl, sGestureName, tGestureTable, iCurFrame, tFrameData, fFrameDelta, fPower)
			end
		end
	end
end

hook.Add("UpdateAnimation", "LuaAnimationSequenceReset", function(pl)
	if pl.InSequence then
		pl:SetSequence("reference")
	end
end)

local meta = _R["Entity"]
function meta:ResetLuaAnimation(sAnimation, fDieTime)
	local animtable = Animations[sAnimation]
	if animtable then
		self.LuaAnimations = self.LuaAnimations or {}
		--[[local desgroup = animtable.Group
		if desgroup then
			for animname, tab in pairs(self.LuaAnimations) do
				if tab.Group == desgroup then
					self.LuaAnimations[animname] = nil
				end
			end
		end]]

		local framedelta = 0
		if animtable.Type == TYPE_POSTURE and not animtable.TimeToArrive then
			framedelta = 1
		end

		if animtable.Type == TYPE_SEQUENCE then
			self.InSequence = true
		end

		self.LuaAnimations[sAnimation] = {Frame = animtable.StartFrame or 1, FrameDelta = framedelta, FrameData = animtable.FrameData, PrevFrameData = animtable.PrevFrameData, Type = animtable.Type, RestartFrame = animtable.RestartFrame, TimeToArrive = animtable.TimeToArrive, Callback = animtable.Callback, ShouldPlay = animtable.ShouldPlay, PreCallback = animtable.PreCallback, Power = animtable.Power or 1, DieTime = fDieTime or animtable.DieTime, Group = animtable.Group}
		self.BuildBonePositions = LuaBuildBonePositions
	end
end

function meta:SetLuaAnimation(sAnimation, fPower)
	if self.LuaAnimations and self.LuaAnimations[sAnimation] then return end

	self:ResetLuaAnimation(sAnimation)
	if fPower then
		self:SetLuaAnimationPower(sAnimation, fPower)
	end
end

function meta:SetLuaAnimationPower(sAnimation, fPower)
	if self.LuaAnimations and self.LuaAnimations[sAnimation] then
		self.LuaAnimations[sAnimation].Power = fPower
	end
end

-- This should be very useful for smoothly stopping an animation instead of instantly stopping it. Looks a lot better than using StopLuaAnimation
function meta:SetLuaAnimationDieTime(sAnimation, fTime)
	if self.LuaAnimations and self.LuaAnimations[sAnimation] then
		self.LuaAnimations[sAnimation].DieTime = fTime
	end
end

function meta:StopLuaAnimation(sAnimation, fTime) -- Time is optional, sets the die time to CurTime() + fTime
	local anims = self.LuaAnimations
	if anims and anims[sAnimation] then
		if fTime then
			anims[sAnimation].DieTime = CurTime() + fTime
		else
			if anims[sAnimation].Type == TYPE_SEQUENCE then
				local count = 0
				for _, tab in pairs(anims) do
					if tab.Type == TYPE_SEQUENCE then
						count = count + 1
					end
				end
				if count <= 1 then
					self.InSequence = nil
				end
			end

			anims[sAnimation] = nil
			if table.Count(anims) <= 0 then
				self.LuaAnimations = nil
				self.BuildBonePositions = nil
			end
		end
	end
end

function meta:StopLuaAnimationGroup(sGroup, fTime)
	if self.LuaAnimations then
		for animname, animtable in pairs(self.LuaAnimations) do
			if animtable.Group == sGroup then
				self:StopLuaAnimation(animname, fTime)
			end
		end
	end
end

function meta:StopAllLuaAnimations(fTime)
	if self.LuaAnimations then
		for name in pairs(self.LuaAnimations) do
			self:StopLuaAnimation(name, fTime)
		end
	end
end
meta = nil
